package com.devexperts.mdd.news.event;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

/**
 * Filter for news events.
 *
 * <p>Sample builder usage:
 * <pre><tt>
 *     NewsFilter filter = NewsFilter.newBuilder()
 *         .withSources("Business Wire")
 *         .withSymbols("AAPL", "MSFT")
 *         .withLimit(50)
 *         .withTimeout(60_000L)
 *         .build();
 * </tt></pre>
 */
public class NewsFilter implements Serializable {

    private static final long serialVersionUID = 1L;

    /** @see NewsFilter#getFeeds() */
    public static final String FILTER_FEED = "feed";
    /** @see NewsFilter#getSources() */
    public static final String FILTER_SOURCE = "source";
    /** @see NewsFilter#getSymbols() */
    public static final String FILTER_SYMBOL = "symbol";
    /** @see NewsFilter#getLimit() */
    public static final String FILTER_LIMIT = "limit";
    /** @see NewsFilter#getTimeout() */
    public static final String FILTER_TIMEOUT = "timeout";

    /** Set of all available fields for the filter. */
    public static final Set<String> FILTER_PARAMETERS = new HashSet<>();
    static {
        FILTER_PARAMETERS.add(NewsFilter.FILTER_FEED);
        FILTER_PARAMETERS.add(NewsFilter.FILTER_SOURCE);
        FILTER_PARAMETERS.add(NewsFilter.FILTER_SYMBOL);
        FILTER_PARAMETERS.add(NewsFilter.FILTER_LIMIT);
        FILTER_PARAMETERS.add(NewsFilter.FILTER_TIMEOUT);
    }

    /** Empty filter with all params set by default. */
    public static final NewsFilter EMPTY_FILTER = newBuilder().build();

    /** Default limit for the news count returned in one chunk. */
    public static final int DEFAULT_LIMIT = 100;

    /** Default time period to wait while polling for news. */
    public static final long DEFAULT_TIMEOUT = 60 * 60 * 1000L; // 1 hour

    // All filter parameters are stored in the map
    private final Map<String, Set<String>> params;

    // Number fields extracted for convenience
    private transient int limit;
    private transient long timeout;

    protected NewsFilter(Map<String, Set<String>> params) {
        this.params = immutableMap(params);
        initialize();
    }

    public static Builder newBuilder() {
        return new Builder();
    }

    public static NewsFilter limitFilter(int limit) {
        return newBuilder().withLimit(limit).build();
    }

    public static NewsFilter emptyFilter() {
        return EMPTY_FILTER;
    }

    // Java Bean API
    
    public Builder builder() {
        return new Builder().withParams(getParams());
    }

    /** Returns set of feeds to filter on, or {@code null} otherwise. */
    public Set<String> getFeeds() {
        return getParam(FILTER_FEED);
    }

    /** Returns set of news sources to filter on, or {@code null} otherwise. */
    public Set<String> getSources() {
        return getParam(FILTER_SOURCE);
    }

    /** Returns set of instrument symbols to filter on, or {@code null} otherwise. */
    public Set<String> getSymbols() {
        return getParam(FILTER_SYMBOL);
    }

    /**
     * Returns maximum number of news to return as requested by client.
     * Note that server can return smaller number than requested due to entitlements.
     */
    public int getLimit() {
        return limit;
    }

    /**
     * Sets the timeout in millis to wait while polling for news.
     * Note that server can use smaller timeout than requested due to entitlements.
     */
    public long getTimeout() {
        return timeout;
    }

    /**
     * Returns all filter parameters as a map.
     */
    public Map<String, Set<String>> getParams() {
        return params;
    }

    public Set<String> getParam(String name) {
        return params.get(name);
    }

    public String toString() {
        return "NewsFilter{" + getParams() + "}";
    }

    // Serialization API

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        initialize();
    }

    // Utility methods

    /** Initialize and validate all required fields. */
    private void initialize() {
        //TODO Remove limits to entitlements
        limit = (int) validNumber(FILTER_LIMIT, getParam(FILTER_LIMIT), DEFAULT_LIMIT);
        timeout = validNumber(FILTER_TIMEOUT, getParam(FILTER_TIMEOUT), DEFAULT_TIMEOUT);
        validateSet(FILTER_SYMBOL, getParam(FILTER_SYMBOL), 10);
        validateSet(FILTER_FEED, getParam(FILTER_FEED), 10);
        validateSet(FILTER_SOURCE, getParam(FILTER_SOURCE), 10);
    }

    private static Map<String, Set<String>> immutableMap(Map<String, Set<String>> map) {
        if (map == null || map.isEmpty())
            return Collections.emptyMap();
        LinkedHashMap<String, Set<String>> copy = new LinkedHashMap<>(map);
        copy.replaceAll((k, values) -> Collections.unmodifiableSet(new LinkedHashSet<>(values)));
        return Collections.unmodifiableMap(copy);
    }

    private static long validNumber(String name, Collection<String> values, long defValue) {
        if (values == null || values.isEmpty()) {
            return defValue;
        }
        long value;
        try {
            value = Long.parseLong(values.iterator().next());
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException("Illegal '" + name + "': value is not a number");
        }
        if (value < 1) {
            throw new IllegalArgumentException("Illegal '" + name + "': " + value + " must be positive");
        }
        return value;
    }

    @SuppressWarnings("SameParameterValue")
    private static void validateSet(String name, Collection<String> values, long maxValue) {
        if (values != null && values.size() > maxValue)
            throw new IllegalArgumentException("Illegal '" + name + "': filter is too large");
    }

    // Builder class for NewsFilter

    public static class Builder {
        private final Map<String, Set<String>> params = new LinkedHashMap<>();

        protected Builder() {}

        // Public API for known parameters

        public Builder withFeeds(String... feeds) {
            return withParam(FILTER_FEED, Arrays.asList(feeds));
        }

        public Builder withFeeds(Collection<String> feeds) {
            return withParam(FILTER_FEED, feeds);
        }

        public Builder withSources(String... sources) {
            return withParam(FILTER_SOURCE, Arrays.asList(sources));
        }

        public Builder withSources(Collection<String> sources) {
            return withParam(FILTER_SOURCE, sources);
        }

        public Builder withSymbols(String... symbols) {
            return withParam(FILTER_SYMBOL, Arrays.asList(symbols));
        }

        public Builder withSymbols(Collection<String> symbols) {
            return withParam(FILTER_SYMBOL, symbols);
        }

        public Builder withLimit(int limit) {
            return withParam(FILTER_LIMIT, String.valueOf(limit));
        }

        public Builder withTimeout(long timeout) {
            return withParam(FILTER_TIMEOUT, String.valueOf(timeout));
        }

        // Generic API methods

        public Builder withParams(Map<String, Set<String>> filters) {
            if (filters != null && !filters.isEmpty()) {
                filters.forEach(this::withParam);
            }
            return this;
        }

        public Builder withParam(String filter, Collection<String> values) {
            if (filter != null && !filter.isEmpty()) {
                if (values != null && !values.isEmpty()) {
                    params.put(filter, new LinkedHashSet<>(values));
                } else {
                    params.remove(filter);
                }
            }
            return this;
        }

        public Builder withParam(String filter, String value) {
            if (filter != null && !filter.isEmpty()) {
                if (value != null && !value.isEmpty()) {
                    params.put(filter, Collections.singleton(value));
                } else {
                    params.remove(filter);
                }
            }
            return this;
        }

        public NewsFilter build() {
            return new NewsFilter(params);
        }
    }
}
